from django.db.models import F
from wbcore import viewsets

from wbcrm.filters.products import ProductCompanyFilterSet, ProductFilterSet
from wbcrm.models import Product, ProductCompanyRelationship
from wbcrm.serializers import (
    ProductCompanyRelationshipModelSerializer,
    ProductModelSerializer,
    ProductRepresentationSerializer,
)
from wbcrm.viewsets.display import ProductCompanyRelationshipDisplay, ProductDisplay
from wbcrm.viewsets.endpoints import ProductCompanyRelationshipEndpointConfig
from wbcrm.viewsets.titles import ProductTitleConfig


class ProductModelViewSet(viewsets.ModelViewSet):
    IDENTIFIER = "wbcrm:product"
    LIST_DOCUMENTATION = "wbcrm/markdown/documentation/product.md"
    queryset = Product.objects.all()
    serializer_class = ProductModelSerializer
    display_config_class = ProductDisplay
    title_config_class = ProductTitleConfig
    search_fields = ("title",)

    filterset_class = ProductFilterSet
    ordering_fields = ("title",)
    ordering = ["title", "is_competitor"]


class ProductCompanyRelationshipCompanyModelViewSet(viewsets.ModelViewSet):
    IDENTIFIER = "wbcrm:company-interestedproduct"
    queryset = ProductCompanyRelationship.objects.all()
    serializer_class = ProductCompanyRelationshipModelSerializer
    display_config_class = ProductCompanyRelationshipDisplay
    title_config_class = ProductTitleConfig
    endpoint_config_class = ProductCompanyRelationshipEndpointConfig
    search_fields = ("product__title",)
    ordering_fields = ("product__title",)
    ordering = ("product__title", "company__name")
    filterset_class = ProductCompanyFilterSet

    def get_queryset(self):
        qs = (
            super()
            .get_queryset()
            .filter(company=self.kwargs["company_id"])
            .annotate(competitor_product=F("product__is_competitor"))
        )
        return qs


class ProductRepresentationViewSet(viewsets.RepresentationViewSet):
    IDENTIFIER = "wbcrm:productrepresentation"
    serializer_class = ProductRepresentationSerializer
    search_fields = ("title",)
    queryset = Product.objects.all()
    ordering = ("title", "is_competitor")



# Generated by Django 4.2.9 on 2024-02-20 10:08

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0007_alter_account_status"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="activity",
            options={"verbose_name": "Activity", "verbose_name_plural": "Activities"},
        ),
    ]



from .accounts import (
    AccountRoleAccountEndpointConfig,
    ChildAccountAccountEndpointConfig,
    InheritedAccountRoleAccountEndpointConfig,
)
from .activities import ActivityEndpointConfig, ActivityParticipantModelEndpointConfig
from .groups import GroupEndpointConfig
from .products import ProductCompanyRelationshipEndpointConfig



from django.apps import apps as global_apps
from django.db import DEFAULT_DB_ALIAS
from django_celery_beat.models import CrontabSchedule, PeriodicTask


def initialize_task(app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
    crontab1, _ = CrontabSchedule.objects.get_or_create(
        minute="0",
        hour="1",
        day_of_week="*",
        day_of_month="*",
        month_of_year="*",
    )
    crontab2, _ = CrontabSchedule.objects.get_or_create(
        minute="0",
        hour="7",
        day_of_week="*",
        day_of_month="*",
        month_of_year="*",
    )

    # Automatically register the utility periodic tasks
    PeriodicTask.objects.update_or_create(
        task="wbcrm.synchronization.activity.tasks.periodic_renew_web_hooks_task",
        defaults={
            "name": "Wbactivity_sync: Renewal of the Activity Sync webhook",
            "crontab": crontab1,
        },
    )
    PeriodicTask.objects.update_or_create(
        task="wbcrm.synchronization.activity.tasks.periodic_notify_admins_of_webhook_inconsistencies_task",
        defaults={
            "name": "Wbactivity_sync: Notification of webhook inconsistencies",
            "crontab": crontab2,
        },
    )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class ActivityEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs):
        user_id = self.request.user.profile.id
        base_url = "wbcrm:activity-list"
        filter_url = f"?participants={user_id}"

        if participants := self.request.GET.get("participants", None):
            filter_url += f",{participants}"
        if companies := self.request.GET.get("companies", None):
            filter_url += f"&companies={companies}"
        return f"{reverse(base_url, args=[], request=self.request)}{filter_url}"

    def get_delete_endpoint(self, **kwargs):
        if "pk" in self.view.kwargs and (self.view.is_private_for_user or self.view.is_confidential_for_user):
            return None
        return super().get_delete_endpoint(**kwargs)


class ActivityParticipantModelEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbcrm:activity-participant-list", args=[self.view.kwargs["activity_id"]], request=self.request)

    def get_create_endpoint(self, **kwargs):
        if "activity_id" in self.view.kwargs:
            return f"{super().get_create_endpoint(**kwargs)}?activity_id={self.view.kwargs['activity_id']}"
        return super().get_create_endpoint(**kwargs)



# Generated by Django 4.1.10 on 2023-08-16 08:16

import django_fsm
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0003_remove_activity_external_id_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="activity",
            name="status",
            field=django_fsm.FSMField(
                choices=[
                    ("CANCELLED", "Cancelled"),
                    ("PLANNED", "Planned"),
                    ("FINISHED", "Finished"),
                    ("REVIEWED", "Reviewed"),
                ],
                default="PLANNED",
                max_length=50,
                verbose_name="Status",
            ),
        ),
    ]



# Generated by Django 5.0.12 on 2025-10-15 13:24

import django.contrib.postgres.search
import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wbcrm', '0017_event'),
    ]

    operations = [
        migrations.AddField(
            model_name='activity',
            name='search_vector',
            field=django.contrib.postgres.search.SearchVectorField(null=True),
        ),
        migrations.AddIndex(
            model_name='activity',
            index=django.contrib.postgres.indexes.GinIndex(fields=['search_vector'], name='activity_sv_gin_idx'),
        ),
    ]



import re
import warnings
from uuid import uuid4

from dateutil.rrule import rrule, rrulestr
from googleapiclient.discovery import Resource
from wbcrm.models import Activity

from ...utils import GoogleSyncUtils


def update_single_event(creator_mail: str, google_service: Resource, internal_activity: Activity, updates: dict):
    """
    Updates a single Google-Event.

    After the external event has been updated, the metadata field of the internal activity is also updated with the information of the external event.

    Note: Do not use for recurring events.

    :param creator_mail:        The e-mail address of the activities creator.
    :param google_service:      The google service to interact with googles resources.
    :param internal_activity:   The internal workbench activity that corresponds to the external google event.
    :param updates:             The update information.
    :returns:                   None
    """

    updated_external_event = (
        google_service.events()
        .update(calendarId=creator_mail, eventId=internal_activity.external_id, body=updates)
        .execute()
    )
    metadata = internal_activity.metadata | {"google_backend": {"event": updated_external_event}}
    Activity.objects.filter(id=internal_activity.id).update(metadata=metadata)


def update_single_recurring_event(
    creator_mail: str, google_service: Resource, internal_activity: Activity, updates: dict
):
    """
    Updates a single recurring Google-Event.

    After the external event has been updated, the metadata field of the internal activity is also updated with the information of the external event.

    Note: Do not use for not recurring events.

    :param creator_mail:        The e-mail address of the activities creator.
    :param google_service:      The google service to interact with googles resources.
    :param internal_activity:   The internal workbench activity that corresponds to the external google event.
    :param updates:             The update information.
    :returns:                   None
    """
    is_parent = Activity.objects.filter(parent_occurrence=internal_activity).exists()
    external_id = (
        internal_activity.metadata["google_backend"]["instance"].get("id")
        if is_parent
        else internal_activity.external_id
    )
    updated_event = google_service.events().patch(calendarId=creator_mail, eventId=external_id, body=updates).execute()
    metadata = internal_activity.metadata
    metadata["google_backend"] |= {"instance": updated_event}
    Activity.objects.filter(id=internal_activity.id).update(metadata=metadata)


def update_all_recurring_events_from_parent(
    creator_mail: str, google_service: Resource, internal_activity: Activity, updates: dict
):
    """
    Updates all Google-Event-Instances belonging to the same recurring event chain.

    After the external event instances have been updated, the metadata fields of the internal activities is also updated with the information of the corresponding external event instance.

    Note: Do not use when creating a new parent activity from an existing child.

    :param creator_mail:        The e-mail address of the activities creator.
    :param google_service:      The google service to interact with googles resources.
    :param internal_activity:   The internal workbench parent activity that corresponds to the external google event.
    :param updates:             The update information.
    :returns:                   None
    """
    updated_event = (
        google_service.events()
        .patch(calendarId=creator_mail, eventId=internal_activity.external_id, body=updates)
        .execute()
    )
    metadata = internal_activity.metadata | {"google_backend": {"event": updated_event}}
    instances = google_service.events().instances(calendarId=creator_mail, eventId=updated_event["id"]).execute()
    google_event_items = instances["items"]
    GoogleSyncUtils.add_instance_metadata(internal_activity, google_event_items, metadata)


def update_all_recurring_events_from_new_parent(
    creator_mail: str, google_service: Resource, internal_activity: Activity, updates: dict
):
    """
    Updates all Google-Event-Instances belonging to the same parent event.

    After the external event instances have been updated, the metadata fields of the internal activities is also updated with the information of the corresponding external event instance.

    Note: Do not use when updating from the original event chain parent.

    :param creator_mail:        The e-mail address of the activities creator.
    :param google_service:      The google service to interact with googles resources.
    :param internal_activity:   The internal workbench parent activity that corresponds to the external google event.
    :param updates:             The update information.
    :returns:                   None
    """

    # If the old parent does not exist anymore, we cannot update the child.
    if not (
        current_parent_occurrence := Activity.objects.filter(
            id=internal_activity.metadata.get("old_parent_id")
        ).first()
    ):
        return warnings.warn(
            "Could not update the recurring events on google, because the old parent activity was already deleted.",
            stacklevel=2,
        )

    # Get the current parent event from google
    current_google_parent_event: dict = (
        google_service.events().get(calendarId=creator_mail, eventId=current_parent_occurrence.external_id).execute()
    )

    # Get the current recurrence rules. We need to modify them for both the current parent event and the new parent event.
    current_parent_rrule_str: str = "\n".join(current_google_parent_event["recurrence"])

    # We need to adjust the rrules to mimic the current status on the workbench. So we replace any until or count value with the current number of child activities.
    current_parent_child_count = Activity.objects.filter(parent_occurrence=current_parent_occurrence).count()
    new_parent_child_count = Activity.objects.filter(parent_occurrence=internal_activity).count()
    current_parent_new_rrule: rrule = rrulestr(current_parent_rrule_str).replace(  # type: ignore
        count=current_parent_child_count + 1, until=None
    )
    new_parent_rrule: rrule = rrulestr(current_parent_rrule_str).replace(count=new_parent_child_count + 1, until=None)  # type: ignore

    # Converting the rrule back to str. Since the .__str__() method adds a DTSTART value, we need to remove this by using regex.
    current_parent_new_rrule_str: str = re.sub("[DTSTART].*[\n]", "", current_parent_new_rrule.__str__()).split("T0")[
        0
    ]
    new_parent_rrule_str: str = re.sub("[DTSTART].*[\n]", "", new_parent_rrule.__str__()).split("T0")[0]

    # Updating the current parent with the new rrules. This will remove all the child events on google that are not in the scope of the changed rrules anymore.
    current_google_parent_event |= {"recurrence": [current_parent_new_rrule_str]}
    updated_current_parent_event: dict = (
        google_service.events()
        .update(calendarId=creator_mail, eventId=current_google_parent_event["id"], body=current_google_parent_event)
        .execute()
    )

    # Updating the corresponding metadata
    current_parent_metadata = current_parent_occurrence.metadata | {
        "google_backend": {"event": updated_current_parent_event}
    }
    Activity.objects.filter(id=current_parent_occurrence.id).update(metadata=current_parent_metadata)
    current_instances = (
        google_service.events()
        .instances(calendarId=creator_mail, eventId=updated_current_parent_event["id"])
        .execute()
    )
    current_instances_google_event_items = current_instances["items"]
    current_parent_occurrence.refresh_from_db()
    GoogleSyncUtils.add_instance_metadata(
        current_parent_occurrence, current_instances_google_event_items, current_parent_metadata
    )

    # Updating the new parent with the created rrules. This will create the child events in google which are in the scope of the newly created rrules.

    internal_activity_query = Activity.objects.filter(id=internal_activity.id)
    external_id = uuid4().hex
    updates |= {"recurrence": [new_parent_rrule_str], "id": external_id}
    internal_activity_query.update(external_id=external_id)
    new_parent_event: dict = google_service.events().insert(calendarId=creator_mail, body=updates).execute()

    # Updating the corresponding metadata
    metadata = internal_activity.metadata | {"google_backend": {"event": new_parent_event}, "old_parent_id": None}
    internal_activity_query.update(metadata=metadata)
    new_instances = (
        google_service.events().instances(calendarId=creator_mail, eventId=new_parent_event["id"]).execute()
    )
    new_google_event_items = new_instances["items"]
    internal_activity.refresh_from_db()
    GoogleSyncUtils.add_instance_metadata(internal_activity, new_google_event_items, metadata, True)



# Generated by Django 5.0.7 on 2024-08-02 13:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0009_alter_account_is_public"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="reference_id",
            field=models.PositiveIntegerField(blank=True, unique=True),
        ),
    ]



from django.utils.translation import gettext as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import BooleanPreference, StringPreference

general = Section("wbactivity_sync")


@global_preferences_registry.register
class BackendCalendarPreference(StringPreference):
    section = general
    name = "sync_backend_calendar"
    default = ""

    verbose_name = _("Synchronization Backend Calendar")
    help_text = _("The Backend Calendar to synchronize activities with an external calendar.")


@global_preferences_registry.register
class SyncPastActivity(BooleanPreference):
    section = general
    name = "sync_past_activity"
    default = False

    verbose_name = _("Synchronization Past Activity")


@global_preferences_registry.register
class SyncCancelledActivity(BooleanPreference):
    section = general
    name = "sync_cancelled_activity"
    default = True

    verbose_name = _("Cancel Internal Activity Instead Of Deleting")
    help_text = _(
        "When an activity is deleted in an external calendar the corresponding workbench activity can be cancelled (default) or also deleted."
    )


@global_preferences_registry.register
class SyncCancelledExternalActivity(BooleanPreference):
    section = general
    name = "sync_cancelled_external_activity"
    default = False

    verbose_name = _("Cancel External Activity With One Non-Attending Internal Participant")
    help_text = _(
        "When an activity was created by an external person and has only one internal participant the activity in the workbench can be canceled if this participant doesn't choose to attend."
    )


@global_preferences_registry.register
class SyncActivityDescription(BooleanPreference):
    section = general
    name = "sync_activity_description"
    default = True

    verbose_name = _("Synchronize Activity Description")


@global_preferences_registry.register
class SyncExternalParticipants(BooleanPreference):
    section = general
    name = "sync_external_participants"
    default = False

    verbose_name = _("Synchronize External Participants From Internal Calendar To External Calendar")


@global_preferences_registry.register
class SyncReplannedReviewedActivityCreatesNewActivity(BooleanPreference):
    section = general
    name = "sync_create_new_activity_on_replanned_reviewed_activity"
    default = False

    verbose_name = _("Create New Activity When Replanning Passed Reviewed Activities")
    help_text = _(
        "If an activity with a past end date (already passed and reviewed) is moved to a future date, a new activity will automatically be created for the updated schedule."
    )


@global_preferences_registry.register
class GoogleSyncCredentials(StringPreference):
    section = general
    name = "google_sync_credentials"
    default = ""
    verbose_name = _("Google Synchronization Credentials")
    help_text = "Dict. Keys: 'url', 'type', 'project_id', 'private_key_id', 'private_key', 'client_email', 'client_id', 'auth_uri', 'token_uri', 'auth_provider_x509_cert_url', 'client_x509_cert_url'"


@global_preferences_registry.register
class OutlookSyncCredentials(StringPreference):
    section = general
    name = "outlook_sync_credentials"
    default = ""
    verbose_name = _("Outlook Synchronization Credentials")
    help_text = '{"notification_url": "", "authority": "", "client_id": "", "client_secret": "", "token_endpoint": "", "graph_url": ""}'


@global_preferences_registry.register
class OutlookSyncAccesToken(StringPreference):
    section = general
    name = "outlook_sync_access_token"
    default = ""

    verbose_name = _("Microsoft Graph Access Token")
    help_text = _("The access token obtained from subscriptions to Microsoft used for authentication pruposes")


@global_preferences_registry.register
class OutlookSyncClientState(StringPreference):
    section = general
    name = "outlook_sync_client_state"
    default = "secretClientValue"

    verbose_name = _("Microsoft Graph Webhook Secret Client State")
    help_text = _(
        "Secret Client Value defined during subscription, it will be injected into the webhook notification against spoofing"
    )



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class GroupEndpointConfig(EndpointViewConfig):
    def get_pre_change_endpoint(self, pk):
        return reverse("wbcrm:group-pre-change", args=[pk], request=self.request)



from collections import OrderedDict
from datetime import date, timedelta

import pandas as pd
from django.contrib.messages import warning
from django.db.models import Q
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from psycopg.types.range import TimestamptzRange
from rest_framework.exceptions import ValidationError
from rest_framework.reverse import reverse
from slugify import slugify
from wbcore import serializers as wb_serializers
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.agenda.serializers import ConferenceRoomRepresentationSerializer
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models import Company, Entry, Person
from wbcore.contrib.directory.preferences import get_main_company
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer
from wbcore.utils.date import calendar_item_shortcuts

from wbcrm.models import Activity, ActivityType
from wbcrm.models.activities import ActivityCompanyThroughModel, ActivityParticipant

from .groups import GroupRepresentationSerializer
from .recurrence import RecurrenceModelSerializerMixin


class ActivityTypeModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = ActivityType
        fields = (
            "id",
            "title",
            "icon",
            "color",
            "score",
            "default",
            "_additional_resources",
        )

    def validate(self, data):
        title = data.get("title", None)
        if title:
            type = ActivityType.objects.filter(slugify_title=slugify(title, separator=" "))
            if self.instance:
                type = type.exclude(id=self.instance.id)
            if type.exists():
                raise ValidationError({"title": _("Cannot add a duplicate activity type.")})

        return data


class ActivityTypeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    endpoint = "wbcrm:activitytyperepresentation-list"
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcrm:activitytype-detail")

    class Meta:
        model = ActivityType
        fields = (
            "id",
            "title",
            "icon",
            "_detail",
            "color",
            "score",
        )


class ActivityRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcrm:activity-detail")
    _detail_preview = wb_serializers.HyperlinkField(reverse_name="wbcrm:activity-detail")
    end__date = wb_serializers.SerializerMethodField()
    title = wb_serializers.CharField(max_length=255, read_only=True)

    label_key = "{{end__date}}: {{title}}"

    def get_end__date(self, obj):
        if obj.end is not None:
            return obj.end.strftime("%d.%m.%Y")
        return ""

    @wb_serializers.register_resource()
    def activity(self, instance, request, user):
        return {"activity": f'{reverse("wbcrm:activity-list")}?participants={instance.id}'}

    class Meta:
        model = Activity
        fields = ("id", "_detail", "_detail_preview", "end", "end__date", "start", "title", "_additional_resources")


def get_default_period(*args, **kwargs):
    return TimestamptzRange(timezone.now(), timezone.now() + timedelta(hours=1))  # type: ignore


def handle_representation(representation: OrderedDict) -> OrderedDict:
    """
    This method is used to remove certain representation values if the representation is for private/confidential use.
    By removing these values, the corresponding fields in the Workbench appear empty.

    :param representation:  Dict of primitive datatypes representing the serializer fields.
    :return:                Either the unchanged representation or a privatized version of the representation
    """
    keys_to_preserve = ["id", "period", "status"]
    private_keys_to_preserve = [
        "participants",
        "_participants",
        "companies",
        "_companies",
        "groups",
        "_groups",
        "assigned_to",
        "_assigned_to",
        "creator",
        "_creator",
    ]
    is_private: bool
    if not ((is_private := representation.get("is_private", False)) or representation.get("is_confidential")):
        return representation
    hidden_representation = OrderedDict.fromkeys(representation, None)
    hidden_representation |= {key: representation.get(key) for key in keys_to_preserve}
    if is_private:
        hidden_representation |= {key: representation.get(key) for key in private_keys_to_preserve}
    hidden_representation["title"] = _("Private Activity") if is_private else _("Confidential Activity")
    return hidden_representation


def _get_default_recurrence_end():
    """
    Default to 6 months in the future
    """
    return (date.today() + pd.tseries.offsets.DateOffset(months=6)).date()


class ActivityCompanyThroughModelRepresentationSerializer(wb_serializers.RepresentationSerializer):
    endpoint = "wbcore:directory:companyrepresentation-list"
    value_key = "id"

    def to_representation(self, value):
        rep = super().to_representation(value)
        rep["id"] = value.company_id
        return rep

    class Meta:
        model = ActivityCompanyThroughModel
        fields = (
            "id",
            # "company",
            "computed_str",
        )


class ActivityModelListSerializer(RecurrenceModelSerializerMixin, wb_serializers.ModelSerializer):
    _companies = ActivityCompanyThroughModelRepresentationSerializer(
        source="activity_companies", related_key="companies", many=True
    )

    _participants = PersonRepresentationSerializer(source="participants", many=True)

    heat = wb_serializers.EmojiRatingField(label="Sentiment")
    _groups = GroupRepresentationSerializer(source="groups", many=True)
    _type = ActivityTypeRepresentationSerializer(source="type")
    _latest_reviewer = PersonRepresentationSerializer(source="latest_reviewer")
    is_private = wb_serializers.BooleanField(default=False, read_only=True)
    is_confidential = wb_serializers.BooleanField(default=False, read_only=True)

    def to_representation(self, instance):
        return handle_representation(super().to_representation(instance))

    def validate(self, data):
        main_company = get_main_company()
        request = self.context["request"]
        companies = data.get("companies", [])
        if main_company and main_company in companies:
            warning(
                request,
                f"The main company {main_company} was removed from the list of companies",
            )
        return data

    class Meta:
        model = Activity
        fields = (
            "id",
            "_additional_resources",
            "type",
            "_type",
            "title",
            "status",
            "period",
            "companies",
            "_companies",
            "participants",
            "_participants",
            "groups",
            "_groups",
            "edited",
            "created",
            "description",
            "result",
            "latest_reviewer",
            "is_private",
            "is_confidential",
            "summary",
            "heat",
            "_latest_reviewer",
        )
        read_only_fields = list(filter(lambda x: x not in ["result"], fields))


class ActivityModelSerializer(ActivityModelListSerializer):
    type = wb_serializers.PrimaryKeyRelatedField(
        queryset=ActivityType.objects.all(), default=lambda: ActivityType.get_default_activity_type()
    )
    repeat_choice = wb_serializers.ChoiceField(
        help_text="Repeat activity at the specified frequency",
        label="Recurrence Frequency",
        required=False,
        read_only=lambda view: not view.new_mode,
        choices=Activity.ReoccuranceChoice.choices,
    )

    recurrence_end = wb_serializers.DateField(
        required=False, label=gettext_lazy("Recurrence End"), default=_get_default_recurrence_end
    )
    visibility = wb_serializers.ChoiceField(
        choices=Activity.Visibility.choices,
        help_text=gettext_lazy(
            "Set to private for the activity to hide sensitive information from anyone but the assignee and participants. Set to confidential to hide from anyone but users with manager permissions."
        ),
        label=gettext_lazy("Visibility"),
    )
    assigned_to = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.CurrentUserDefault("profile"),
        queryset=Person.objects.all(),
        label=gettext_lazy("Assigned to"),
    )
    _assigned_to = PersonRepresentationSerializer(source="assigned_to")
    creator = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.CurrentUserDefault("profile"),
        many=False,
        read_only=True,
    )
    _creator = PersonRepresentationSerializer(
        source="creator",
    )
    companies = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.DefaultFromGET("companies", many=True),
        many=True,
        queryset=Company.objects.all(),
        label=gettext_lazy("Companies"),
    )
    participants = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.DefaultFromGET("participants", many=True),
        many=True,
        queryset=Person.objects.all(),
        label=gettext_lazy("Participants"),
    )
    period = wb_serializers.DateTimeRangeField(
        label=gettext_lazy("Period"),
        default=get_default_period,
        shortcuts=calendar_item_shortcuts,
        read_only=lambda view: view.is_external_activity,
    )

    is_private = wb_serializers.BooleanField(default=False, read_only=True)
    is_confidential = wb_serializers.BooleanField(default=False, read_only=True)
    _conference_room = ConferenceRoomRepresentationSerializer(source="conference_room")

    title = wb_serializers.CharField(
        default=wb_serializers.DefaultFromGET("title"),
        label=gettext_lazy("Title"),
        read_only=lambda view: view.is_external_activity,
    )
    all_day = wb_serializers.BooleanField(read_only=lambda view: view.is_external_activity)
    description = wb_serializers.TextField(read_only=lambda view: view.is_external_activity)

    @cached_property
    def user(self) -> User | None:
        if request := self.context.get("request"):
            return request.user
        return None

    @wb_serializers.register_resource()
    def activity_participants_table(self, instance, request, user):
        return {
            "activity_participants_table": reverse(
                "wbcrm:activity-participant-list", kwargs={"activity_id": instance.id}, request=request
            )
        }

    def to_representation(self, instance):
        return handle_representation(super().to_representation(instance))

    def validate(self, data):  # noqa: C901
        if (
            (result := data.get("result", None))
            and result not in ["", "<p></p>"]
            and result != getattr(self.instance, "result", "")
        ):
            data["latest_reviewer"] = self.user.profile
            data["reviewed_at"] = timezone.now()
        if data.get("creator", None) is None:
            data["creator"] = self.user.profile

        if not data.get("title", self.instance.title if self.instance else None):
            raise ValidationError({"title": "You need to specify a title for this activity."})

        period = data.get("period", self.instance.period if self.instance else None)
        if not data.get("type", self.instance.type if self.instance else None):
            raise ValidationError({"type": _("Please add an activity type.")})

        if room := data.get("conference_room", self.instance.conference_room if self.instance else None):
            qs = Activity.objects.filter(status=Activity.Status.PLANNED, conference_room=room)
            if self.instance:
                qs = qs.exclude(id=self.instance.id)
            conference_room_activity = Activity.get_inrange_activities(
                qs, period.lower + timedelta(seconds=60), period.upper - timedelta(seconds=60)
            ).first()
            if conference_room_activity and (user := self.user):
                if (
                    conference_room_activity.visibility == Activity.Visibility.PRIVATE
                    and user.profile not in conference_room_activity.participants.all()
                    and user.profile != conference_room_activity.assigned_to
                ):
                    raise ValidationError(
                        {
                            "non_field_errors": _(
                                "A private activity already uses this conference room at the same time."
                            )
                        }
                    )
                elif (
                    conference_room_activity.visibility == Activity.Visibility.CONFIDENTIAL
                    and not CalendarItem.has_user_administrate_permission(user)
                ):
                    raise ValidationError(
                        {
                            "non_field_errors": _(
                                "A confidential activity already uses this conference room at the same time."
                            )
                        }
                    )
                else:
                    raise ValidationError(
                        {
                            "non_field_errors": _(
                                'The activity "{title}" already uses this conference room at the same time.'
                            ).format(title=conference_room_activity.title)
                        }
                    )

        # In the following section we validate the groups and the group members. If a user trys to remove a group member from the participants/companies fields, we want to raise an validation error.
        if self.instance and self.instance.groups.exists() and not ("groups" in data and data["groups"] == []):
            # If there are no companies or participants in the current instance, you cannot remove a group member; therefore there is no need to validate the groups. This can happen if you add to an existing and empty activity a group.
            if self.instance.companies.exists() or self.instance.participants.exists():
                companies = data.get("companies", [])
                participants = data.get("participants", [])
                # If there are new companies or participants in data, we need to check if existing group members have been removed.
                if "companies" in data or "participants" in data:
                    new_groups = data.get("groups", [])
                    new_groups_id_list = [group.id for group in new_groups]
                    old_groups = self.instance.groups.exclude(id__in=new_groups_id_list)

                    # If there are no new groups in data, we need to check if members of the old group have been removed.
                    # If there are new groups and also old groups, we need to check for the old groups if members have been removed.
                    # If only new groups are present, we don't need to validate.
                    if not new_groups or old_groups.exists():
                        new_companies_id_list = (
                            [companies.id for companies in companies]
                            if "companies" in data
                            else self.instance.companies.values_list("id", flat=True)
                        )
                        new_participants_id_list = (
                            [participant.id for participant in participants]
                            if "participants" in data
                            else self.instance.participants.values_list("id", flat=True)
                        )
                        new_entries = Entry.objects.filter(
                            Q(id__in=new_companies_id_list) | Q(id__in=new_participants_id_list)
                        ).distinct()
                        missing_members = (
                            Entry.objects.filter(groups__in=old_groups).exclude(id__in=new_entries).distinct()
                        )
                        # If there are group members who are not in the new entries field, we can assume, that a group member has been removed and we need to throw an error.
                        if missing_members.exists():
                            missing_members_computed_strs = ", ".join(
                                list(missing_members.values_list("computed_str", flat=True))
                            )
                            util_str = _("is a member")
                            if missing_members.count() > 1:
                                start, x, end = missing_members_computed_strs.rpartition(",")
                                missing_members_computed_strs = start + _(" and") + end
                                util_str = _("are members")

                            groups = ", ".join(
                                list(dict.fromkeys(missing_members.values_list("groups__title", flat=True)))
                            )
                            error_message = _(
                                "{missing_members} {util_str} of the following group(s): {groups}\n. You cannot remove members of selected groups.",
                            ).format(missing_members=missing_members_computed_strs, util_str=util_str, groups=groups)

                            raise ValidationError({"groups": error_message})
        return super().validate(data)

    class Meta:
        model = Activity
        required_fields = ("importance", "period", "reminder_choice", "repeat_choice", "start", "title", "type")
        read_only_fields = (
            "creator",
            "edited",
            "latest_reviewer",
            "reviewed_at",
            "is_cancelled",
        )
        fields = (
            "id",
            "_additional_resources",
            "all_day",
            "assigned_to",
            "_assigned_to",
            "companies",
            "_companies",
            "conference_room",
            "_conference_room",
            "created",
            "creator",
            "_creator",
            "description",
            "disable_participant_check",
            "edited",
            "end",
            "groups",
            "_groups",
            "importance",
            "visibility",
            "latest_reviewer",
            "_latest_reviewer",
            "location",
            "location_latitude",
            "location_longitude",
            "online_meeting",
            "participants",
            "_participants",
            "summary",
            "period",
            "propagate_for_all_children",
            "recurrence_count",
            "recurrence_end",
            "reminder_choice",
            "repeat_choice",
            "result",
            "reviewed_at",
            "start",
            "status",
            "title",
            "type",
            "_type",
            "is_cancelled",
            "is_private",
            "is_confidential",
            "_buttons",
        )


class ReadOnlyActivityModelSerializer(ActivityModelSerializer):
    class Meta(ActivityModelSerializer.Meta):
        read_only_fields = list(filter(lambda x: x not in ["result"], ActivityModelSerializer.Meta.fields))


class ActivityParticipantModelSerializer(wb_serializers.ModelSerializer):
    _activity = ActivityRepresentationSerializer(source="activity")
    activity = wb_serializers.PrimaryKeyRelatedField(
        default=wb_serializers.DefaultFromGET("activity_id"),
        queryset=Activity.objects.all(),
        label=gettext_lazy("Activity"),
    )
    customer_status = wb_serializers.CharField(
        default="", label=gettext_lazy("Status"), allow_null=True, read_only=True
    )
    position = wb_serializers.CharField(default="", label=gettext_lazy("Position"), read_only=True)
    primary_telephone = wb_serializers.TelephoneField(
        default="", label=gettext_lazy("Primary Telephone"), allow_null=True, read_only=True
    )
    primary_email = wb_serializers.CharField(
        allow_null=True, default="", label=gettext_lazy("Primary Email"), read_only=True
    )
    _participant = PersonRepresentationSerializer(source="participant")
    is_occupied = wb_serializers.BooleanField(default=False)

    def validate(self, data):
        data_activity = data.get("activity", None)
        data_participant = data.get("participant", None)

        if ActivityParticipant.objects.filter(
            activity=data_activity,
            participant=data_participant,
        ).exists():
            raise ValidationError({"participant": _("The person is already a participant in this activity.")})

        return super().validate(data)

    def create(self, validated_data):
        validated_data.pop("is_occupied", None)
        return super().create(validated_data)

    class Meta:
        model = ActivityParticipant
        fields = (
            "_activity",
            "_additional_resources",
            "_participant",
            "activity",
            "customer_status",
            "position",
            "id",
            "is_occupied",
            "participant",
            "participation_status",
            "status_changed",
            "primary_telephone",
            "primary_email",
        )



# Generated by Django 5.0.9 on 2024-10-18 13:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0011_activity_summary"),
    ]

    operations = [
        migrations.AlterField(
            model_name="activity",
            name="summary",
            field=models.TextField(blank=True, default="", verbose_name="LLM Summary"),
        ),
    ]
